RW = LibStub("AceAddon-3.0"):NewAddon("Raid Watch", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
if not RW then return end

--[===[@debug@
RW.debug = true
--@end-debug@]===]

-- Lua funcs
local pairs, tonumber, error, type = pairs, tonumber, error, type
local tostring = tostring

-- Version
RW.RELEASETYPE = "Release"
RW.VERSION = GetAddOnMetadata("RW_Core", "Version") or "1.0"
RW.REVISION = tonumber("255")
if not RW.REVISION then 
	RW.REVISION = "Dev" 
	RW.RELEASETYPE = "Development"
end

if RW.RELEASETYPE == "Release" then
	--@alpha@
	RW.RELEASETYPE = "Alpha"
	--@end-alpha@
end

local UIFSettings = {
		s = {
			font = LibStub("LibSharedMedia-3.0"):Fetch("font", "accid"),
			fontSize = 16,
		}
	}
	
local UIF = LibStub("AceGUI-Factory-1.0"):GetFactory("RW", UIFSettings)
UIF:SetValuesChangedCallback(function()
	RW:SendMessage("RW_GUIValuesChanged")
end)

-- Bits used to control usage of plugins (Inspired by Bigwigs)
RW.BITS = {
	BAR 	= 0x00000001,
	MESSAGE = 0x00000002,
	WARNING = 0x00000004,
	ICON 	= 0x00000008,
	RANGE 	= 0x00000010,
}

-- Spellname lookup
RW.SpellName = setmetatable({}, {
	__index = function(t, key)
		if type(key) ~= "number" then return key end
		local name = GetSpellInfo(key)
		if not name then error(("Cannot fetch spell name for %q"):format(key)) end
		return name		
	end}
)

function RW:OnInitialize()
	local defaults = {profile = {
		minimap = {
			hide = false,
		},
		vicotorySound = "RW: Victory"
	}}
	self.db = LibStub("AceDB-3.0"):New("RaidWatch2DB", defaults, "Default")
	self:SetUpLDB()
	self.db.RegisterCallback(self, "OnProfileChanged", "Reset")
	self.db.RegisterCallback(self, "OnProfileCopied", "Reset")
	self.db.RegisterCallback(self, "OnProfileReset", "Reset")
	self.modStorage = {}
	self.bossStorage = {}
	self.iconStorage = {}
	
	self:RegisterMessage("RW_BossModFound")
	self:RegisterMessage("RW_BossRegistered")
end

function RW:OnEnable()
	self:RegisterMessage("RW_BossModFound")
	self:RegisterMessage("RW_BossRegistered")
	
	self:RegisterChatCommand("mrw", "SlashHandler")
	self:RegisterChatCommand("rw2", "SlashHandler")
	self:RegisterChatCommand("range", "ShowRange")
	
	if IsInGuild() then
		self:ScheduleTimer("SendVersion", 4)
		self:SendMessage("RW_RegisterCommListener", self, "IncGuildVersionInfo")
		self.GuildVersions = {}
	end
end

function RW:OnDisable()
	self:SendMessage("RW_UnregisterAllCommListeners", self)
end

function RW:Reset()
	local icon = LibStub("LibDBIcon-1.0")
	if not icon then return end
	
	if RW.db.profile.minimap.hide then
		icon:Hide("RW2")
	else
		icon:Show("RW2")
	end
end

function RW:SendVersion()
	self:SendMessage("RW_SendComm", "IncGuildVersionInfo", "GUILD", nil, RW.VERSION, RW.REVISION, RW.RELEASETYPE)
end

function RW:IncGuildVersionInfo(sender, dist, version, revision, releaseType)
	
	self.GuildVersions[sender] = self.GuildVersions[sender] or {}
	self.GuildVersions[sender].name = sender
	self.GuildVersions[sender].version = version
	self.GuildVersions[sender].revision = revision
	self.GuildVersions[sender].releaseType = releaseType
	
	if sender ~= UnitName("player") and dist == "GUILD" then
		self:SendMessage("RW_SendComm", "IncGuildVersionInfo", "WHISPER", sender, RW.VERSION, RW.REVISION, RW.RELEASETYPE)
	end
end

-- General message function, usualy embedded in plugins 
-- and bosses to automaticly supply their own prefix.
function RW:Msg(msg, prefix)
	local prefix = prefix or self:GetName()
	DEFAULT_CHAT_FRAME:AddMessage(("|cff28579d[|r|cffcfe846%s|r|cff28579d]|r "):format(tostring(prefix))..msg, 0.39, 0.67, 0.87)
end

---------------------------------------------
----- Modules handling 
---------------------------------------------

-- Plugins
RW.plugins = RW:NewModule("Plugins")
RW.plugins:SetDefaultModuleState(true)
RW.plugins:SetDefaultModuleLibraries("AceEvent-3.0", "AceTimer-3.0")
--- Register a plugin
-- @param name The name of the module to register
-- @returns The created module
function RW:Plugin(name, ...)
	local module = RW.plugins:NewModule(name, ...)
	module.Msg = self.Msg
	return module
end

-- Bosses
RW.bosses = RW:NewModule("Bosses")
RW.bosses:SetDefaultModuleState(false)
RW.bosses:SetDefaultModuleLibraries("AceEvent-3.0", "AceTimer-3.0")
--- Register a boss module
-- @param name Name of the boss
-- @param loadZone Zone/Zones where this boss module should be enabled/tracking
-- @param modPack The mod package this boss belongs to
-- @param subGroup If passed it will be used in the options to sub group (Wings in Naxx for example)
-- @param ... Default libraries to embed for the boss
-- @return The boss module
function RW:Boss(name, loadZone, modPack, subGroup, icon, ...)
	local boss = RW.bosses:NewModule(name, ...)
	boss.modPack = modPack
	boss.loadZone = (type(loadZone) == "table") and loadZone or {loadZone}
	boss.Msg = self.Msg
	boss.subGroup = subGroup
	if icon then
		boss.icon = icon
	end
	
	boss.type = "raid" -- Default to raid
	
	return boss
end

function RW:GetPlugin(name)	return RW.plugins:GetModule(name, true) end
function RW:IteratePlugins() return RW.plugins:IterateModules() end
function RW:GetBoss(name) return RW.bosses:GetModule(name, true) end
function RW:IterateBosses() return RW.bosses:IterateModules() end

function RW:RegisterBoss(boss)
	
	-- For svn version
	--print(self.modStorage)
	if not self.modStorage[boss.modPack] then
		self:SendMessage("RW_BossModFound", boss.modPack)
		self:SendMessage("RW_NewMonitorZone", boss.loadZone)
	end
	
	if boss.OnRegister then boss:OnRegister() end
	
	boss.db = self.db:RegisterNamespace(boss:GetName(), {profile = boss.defaults })	
	
	-- Erase old data
	boss.db.profile.Stats = nil
	
	self:SendMessage("RW_BossRegistered", boss)
	boss.loadZone = nil
end

function RW:RW_BossModFound(message, name, id, icon)
	self.modStorage[name] = id
	self.iconStorage[name] = icon
end

function RW:RW_BossRegistered(message, boss)
	tinsert(self.bossStorage, boss)
end

function RW:RegisterPlugin(plugin)
	if plugin.OnRegister then plugin:OnRegister() end
	self:SendMessage("RW_PluginRegistered", plugin)
end

--------------------------------------------
-- Slash commands

function RW:SlashHandler(input)
	if input == "" or input == "config" then
		local config = self:GetPlugin("Options")
		if not config then	
			LoadAddOn("RW_Options")
			if IsAddOnLoaded("RW_Options") then
				config = self:GetPlugin("Options")
			else 
				self:Msg("Could not open options, |cFF10FF32RW_Options|r could not be found")
			end
		end
		if config then config:Toggle() end
	end
end

function RW:ShowRange()
	self:SendMessage("RW_ShowRangeCheck", true, 11)
end
---------------------------------------------
-- LDB

function RW:SetUpLDB()
	local LDB = LibStub("LibDataBroker-1.1")
	if not LDB then return end
	local l = LDB:NewDataObject("RW2")
	l.type = "launcher"
	l.icon = [[Interface\Icons\Achievement_Dungeon_NexusRaid_Normal.blp]]
	l.tocname = "RW_Core"
	l.label = "Raid Watch 2"
	l.OnClick = function()
		RW:SlashHandler("config")
	end
	l.OnTooltipShow = function(tt)
		tt:AddLine("|cffcfe846Raid Watch 2|r")
		tt:AddLine("Click to show settings")
	end
	local LibDBIcon = LibStub("LibDBIcon-1.0")
	if not LibDBIcon then return end
	LibDBIcon:Register("RW2", l, self.db.profile.minimap)
end
