local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0")

local AL = LibStub("AceLocale-3.0")
local LSM = LibStub("LibSharedMedia-3.0")

local Boss = {}
RW.bossPrototype = Boss
RW.bosses:SetDefaultModulePrototype(Boss)

-- Party module, have all the info about the current party
local Party
local Comm

function Boss:OnInitialize()
	self.defaults = {
		enabled = true, 
		SortBossCategories = false,
	}
	self.BaseAlertObjects = {}
	self.AlertObjects = {}
	
	-- Some constants
	self.pName = UnitName("player")
	self.pClass = select(2, UnitClass("player"))
	self.isCaster = self.pClass == "MAGE" or self.pClass == "WARLOCK" or self.pClass == "DRUID" or self.pClass == "PRIEST" or self.pClass == "SHAMAN"
	self.isMelee = self.pClass == "WARRIOR" or self.pClass == "DEATHKNIGHT" or self.pClass == "ROGUE" or self.pClass == "DRUID" or self.pClass == "SHAMAN"
	
	RW:RegisterBoss(self) -- Calls OnRegister for the boss module
	
	self.db.RegisterCallback(self, "OnProfileChanged", "Reset")
	self.db.RegisterCallback(self, "OnProfileCopied", "Reset")
	self.db.RegisterCallback(self, "OnProfileReset", "Reset")
	self.DB = self.db.profile
	
	Party = RW:GetPlugin("Party")
	Comm = RW:GetPlugin("Comm")
	
	self.Party = Party
	
	-- Clean up after init
	--self.defaults = nil
end

function Boss:OnEnable()
	self:RegisterMessage("RW_BarFinished")
end

function Boss:OnDisable()
	-- Cleanup
	self.engageTime = nil
	if self.trackedSpells then wipe(self.trackedSpells) end
	if self.trackedYells then wipe(self.trackedYells) end
	if self.trackedDeaths then wipe(self.trackedDeaths) end
	
	-- Stop all alert objects and hide them
	self:UnscheduleAll()
	self:StopAllTimers()
	self:ResetAllIcons()
	self:SendMessage("RW_RemoveAllHealthFrames")
	self:SendMessage("RW_ShowRangeCheck", false)
	self:SendMessage("RW_HideAllArrows")
end

function Boss:Reset()
	self.DB = self.db.profile
end

function Boss:Start(trigger)
	if not self.TrashMod then
		if trigger then
			RW:Msg(L.BOSS_ENGAGE:format(rawget(self.L, "name") or boss:GetName()))
			self:SendMessage("RW_Warning", L.BOSS_ENGAGE:format(rawget(self.L, "name") or self:GetName()), self.BossIcon, nil, "None")
		end
	end
	self.engageTime = GetTime()
	
	self:Enable()
	if self.OnEngage then self:OnEngage(trigger) end
end

function Boss:End(wipe, trigger)
	if not self.TrashMod then
		local timeTaken
		if self.engageTime then
			timeTaken = GetTime() - self.engageTime
		else
			timeTaken = 0
		end

		if trigger then
			-- Did not wipe
			if not wipe then
				RW:Msg(L.BOSS_WIN:format(self:GetName(), RW:TimeToStr(timeTaken)))
				PlaySoundFile(LSM:Fetch("sound", RW.db.profile.vicotorySound))
				self:SendMessage("RW_Warning", (rawget(self.L, "name").." defeated!" or self:GetName()), self.BossIcon, nil, "None")				
			-- Wipe
			else
				RW:Msg(L.BOSS_WIPE:format(self:GetName(), RW:TimeToStr(timeTaken)))
			end
		end
	end
	
	self:Disable()
	if self.OnEnd then self:OnEnd(wipe, trigger) end
end


function Boss:RegisterLocale(lang)
	return AL:NewLocale(self:GetName(), lang, true)
end

function Boss:GetLocale()
	--@non-debug@
	return AL:GetLocale(self:GetName(), true)
	--@end-non-debug@
	--[===[@debug@
	return AL:GetLocale(self:GetName())
	--@end-debug@]===]
end

function Boss:SetAsTrashMod()
	self.TrashMod = true
end

function Boss:GetOpt(optName)
	return self.db.profile[optName]
end

function Boss:SetDelay(delay)
	self.endDelay = delay
end

function Boss:SetTrigger(...)
	for i = 1, select("#", ...) do
		local v = select(i, ...)
		if type(v) == "string" then
			self.yellTriggers = self.yellTriggers or {}
			self.yellTriggers[v] = true
		elseif type(v) == "number" then
			self.mobTriggers = self.mobTriggers or {}
			self.mobTriggers[v] = true
		end
	end
end

function Boss:SetWin(...)
	if not ... then
		self.mobWins = self.mobWins or {}
		self.mobWins[self:GetFirstTrigger()] = true
	elseif select(1, ...) == "combat" then
		self.winTrigger = "combat"
	end
	for i = 1, select("#", ...) do
		local v = select(i, ...)
		if type(v) == "string" then
			self.yellWins = self.yellWins or {}
			self.yellWins[v] = true
		elseif type(v) == "number" then
			self.mobWins = self.mobWins or {}
			self.mobWins[v] = true
		end
	end
end

function Boss:GetFirstTrigger()
	local cID
	for mob in pairs(self.mobTriggers) do
		cID = mob
		break
	end
	return cID
end

-------------------------------------
-- Party functions
function Boss:GetPartyType()
	return Party:GetPartyType()
end

function Boss:IsHeroic()
	if self.type == "raid" then
		local diff = GetRaidDifficulty()
		if diff == 3 or diff == 4 then
			return true
		else 
			return false
		end
	elseif self.type == "party" then
		local diff = GetDungeonDifficulty()
		if diff == 2 then 
			return true
		else 
			return false
		end
	end
end

function Boss:InstanceSize()
	local size
	if self.type == "raid" then
		local diff = GetRaidDifficulty()
		if diff == 1 or diff == 3 
			then size = 10
		else
			size = 25
		end
	elseif self.type == "party" then
		size = 5
	end
	
	return size
end

function Boss:IsInstanceType(...)
	local name, instanceType, difficultyIndex, difficultyName, maxPlayers = GetInstanceInfo()
	local iString = (self:IsHeroic() and "heroic" or "normal")..maxPlayers
	for i = 1, select("#", ...) do
		if iString == slect(i, ...) then return true end
	end
	
	return false
end

-------------------------------------
-- cID functions
function Boss:CUIDFromGUID(gUID)
	if not gUID then return end
	return tonumber(gUID:sub(9, 12), 16)
end

function Boss:CUIDFromUID(uID)
	if not uID then return end
	local gUID = UnitGUID(uID)
	if gUID then
		return self:CUIDFromGUID(gUID)
	end
end

-------------------------------------
-- uID Functions
function Boss:UIDFromCID(cID)
	if self:CUIDFromGUID(UnitGUID("mouseover")) == cID then
		return "mouseover"
	end
	for name, member in pairs(Party.group) do
		if self:CUIDFromGUID(UnitGUID(member.id.."target")) == cID then
			return member.id.."target"
		end
	end
end

function Boss:UIDFromGUID(gUID)
	if not gUID then return end
	if UnitGUID("mouseover") == gUID then
		return "mouseover"
	end
	for name, member in pairs(Party.group) do
		if UnitGUID(member.id.."target") == gUID then
			return member.id.."target"
		end		
	end
end

function Boss:UIDFromName(name)
	if UnitName("mouseover") == name then
		return "mouseover"
	end
	for name, member in pairs(Party.group) do
		if UnitName(member.id.."target") == name then
			return member.id.."target"
		end
	end
end

function Boss:UIDFromRandom(target)
	local targetID
	if target:find("^0x") then		-- Target by gUID
		targetID = self:UIDFromGUID(target)
	elseif Party.group[target] then	-- Target by party member name
		targetID = Party.group[target].id
	else		
		targetID = self:UIDFromName(target)  -- Target by mob name
	end
	return targetID
end

-- Target
function Boss:GetMobTarget(cID)
	if not cID then cID = self:GetFirstTrigger() end
	if not cID then return end
	if cID == Boss:CUIDFromUID("target") then return UnitName("targettarget") end
	for i, player in pairs(Party.group) do
		if cID == Boss:CUIDFromUID(player.id.."target") then return UnitName(player.id.."targettarget") end
	end
end

-- Health
function Boss:GetBossHealth(cID)
	cID = cID or self:GetFirstTrigger()
	if not cID then return 0.0 end
	local uID = self:UIDFromCID(cID)
	if not uID then return 0.0 end
	return UnitHealth(uID) / UnitHealthMax(uID)
end

function Boss:SetIcon(icon, target)
	SetRaidTarget(target, icon or 0)
end

-------------------------------------
-- Tacking

function Boss:Track(func, ...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	self:SendMessage("RW_TrackRegistered", self, func, ...)
	self.trackedSpells = self.trackedSpells or {}
	for i = 1, select("#", ...) do
		self.trackedSpells[select(i, ...)] = func
	end
end

function Boss:YellTrack(func, ...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	self:SendMessage("RW_YellRegistered", self, func, ...)
	self.trackedYells = self.trackedYells or {}
	for i = 1, select("#", ...) do
		self.trackedYells[select(i, ...)] = func
	end
end

function Boss:DeathTrack(func, ...)
	if type("func") == "string" and not self[func] then
		error(("Missing function %q for bossmod %q"):format(func, self:GetName()))
	end
	self:SendMessage("RW_DeathRegistered", self, func, ...)
	self.trackedDeaths = self.trackedDeaths or {}
	for i = 1, select("#", ...) do
		self.trackedDeaths[select(i, ...)] = func
	end
end


-------------------------------------
-- Scheduling

do
	local timers = {}
	local args = {}
		
	local function clearTimer(id)
		if not id or not timers[id] then return end
		local timer = timers[id]
		if timer.args then wipe(timer.args) end
		wipe(timer)
		timers[id] = nil
	end
	
	local function handleTimers(id)
		local timer = timers[id]
		local func = timer.func
		local m = timer.module
		if timer.args then
			wipe(args)
			for i, v in pairs(timer.args) do tinsert(args, v) end
		end
		if type(timer.func) == "function" then
			if not timer.rep then
				clearTimer(id)	
			end
			func(unpack(args))
		else
			if not timer.rep then
				clearTimer(id)	
			end
			if m[func] then
				m[func](m, unpack(args))
			end
		end
	end
	
	local function prepareSchedule(self, id, time, func, ...)
		local timer = timers[id]
		if timer then
			self:Unschedule(id)
		else
			timer = {}
		end
		timer.func = func
		timer.module = self
		timer.args = {}
		if select("#", ...) > 0 then
			if timer.args then
				for i = 1, select("#", ...) do
					tinsert(timer.args, (select(i, ...)))
				end
			else
				timer.args = {...}
			end
		end
		timers[id] = timer
	end
	
	function Boss:Schedule(id, time, func, ...)
		prepareSchedule(self, id, time, func, ...)
		timers[id].id = self:ScheduleTimer(handleTimers, time, id)
	end
	
	function Boss:ScheduleRepeating(id, time, func, now, ...)
		prepareSchedule(self, id, time, func, ...)
		timers[id].rep = true
		timers[id].id = self:ScheduleRepeatingTimer(handleTimers, time, id)
		if now then handleTimers(id) end
	end

	function Boss:Unschedule(id)
		if timers[id] then
			self:CancelTimer(timers[id].id, true)
			clearTimer(id)
		end
	end

	function Boss:UnscheduleAll()
		for id, timer in pairs(timers) do
			if timer.module == self then
				self:Unschedule(id)
			end
		end
	end
end

function Boss:AddBoolOption(varName, optionName, category, default, icon)
	base = {
		type = "bool",
		icon = icon,			
		id = optionName,
		var = varName,
		mod = self,
	}
	self.defaults[varName] = ((default == nil) or default)
	tinsert(self.BaseAlertObjects, base)
	self:AddToCategory(category, optionName)
end

function Boss:AddDefaults(optionName, aType, sType, default)
	self.defaults[optionName] = self.defaults[optionName] or {}	
	self.defaults[optionName][aType] = self.defaults[optionName][aType] or {}
	self.defaults[optionName][aType][sType] = {
		enabled = ((default == nil) or default)
	}
	self.defaults[optionName].enabled = true
	
	return self.defaults[optionName][aType][sType]
end

do
	local function AddToGeneral(self, ...)
		local genCat
		if not RW:InTable2(self.Options.Categories, "text", "General") then
			genCat = {text = "General"}
			tinsert(self.Options.Categories, genCat)
		else
			genCat = RW:GetTableData(self.Options.Categories, "text", "General")
		end
		for i=1, select("#", ...) do
			local used = false
			local newopt = select(i, ...)
			for _, cat in pairs(self.Options.Categories) do
				for j, opt in pairs(cat) do
					if opt == newopt then used = true end
				end
			end
			if genCat then
				if not used then tinsert(genCat, newopt) end
			end
		end
	end

	function Boss:AddToCategory(cat, ...)
		self.Options = self.Options or {}
		self.Options.Categories = self.Options.Categories or {}
		
		if not cat then return AddToGeneral(self, ...) end
			
		local newCat
		if RW:InTable2(self.Options.Categories, "text", cat) then
			for _, oldcat in pairs(self.Options.Categories) do				
				if oldcat.text == cat then
					newCat = oldcat
				end
			end
		else
			newCat = {
				text = cat
			}
			tinsert(self.Options.Categories, newCat)
		end		
		
		for i=1, select("#", ...) do
			local opt = select(i, ...)
			tinsert(newCat,  type(opt) == "number" and RW.SpellName[opt] or opt)
		end
	end
end

function Boss:GetBaseAlertObject(id)
	for i, obj in pairs(self.BaseAlertObjects) do
		if obj.id == id then return obj end
	end
end

function Boss:GetIconObject(id)
	if not self.AlertObjects.ICON then return end
	for i, obj in pairs(self.AlertObjects.ICON) do
		if id == obj.base.id then return obj end
	end
end
